/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include <sys/socket.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "monitor.h"


/**
 *  Manage process information persistently. Process data is saved to
 *  a state file (~/.monit.state) when monit runs in daemon mode for
 *  each poll cycle. Monit use this file to recover from a crash or to
 *  maintain process data persistently during a reload.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @version \$Id: state.c,v 1.4 2003/04/23 08:46:46 martinp Exp $
 *  @file
 */


/* ------------------------------------------------------------- Definitions */


/**
 * Fields from the Process_T object type, which we are interessted in
 * when handling the state.
 */
typedef struct mystate {
  char name[STRLEN];   
  int  mode;       
  int  nstart;     
  int  ncycle;     
  int  do_validate;
  int  has_checksum_error;
} State_T;


/* -------------------------------------------------------------- Prototypes */


static void close_state(FILE *);
static FILE *open_state(const char *mode);
static void clone_state(Process_T, State_T *);
static void update_process_state(Process_T, State_T *);


/* ------------------------------------------------------------------ Public */



/**
 * Save process state information to the state file
 */
void state_save() {

  int l= 0;
  State_T s;
  Process_T p;
  FILE *S= NULL;
  
  if(! (S= open_state("w")))
      return;

  l= get_process_list_length();
  
  if(fwrite(&l, 1, sizeof (int), S) != sizeof(int)) {
    log("%s: Unable to save monit state information to '%s'\n",
	prog, Run.statefile);
    goto error;
  }
  
  for(p= processlist; p; p= p->next) {
    clone_state(p, &s);
    if(fwrite(&s, 1, sizeof(State_T), S) != sizeof(State_T)) {
      log("%s: An error occured when saving monit state information "
	  "for the process %s\n", prog, p->name);
      goto error;
    }
  }
  
  error:
  close_state(S);
  
}


/**
 * Check if we should update current processes with persistent state
 * information. The logic is as follows: Iff a state file is present
 * *and* older than the running monit daemon's lock file we have had a
 * crash and should update data from the state file.
 * @return TRUE if the state should be updated otherwise FALSE
 */
int state_should_update() {

  if(exist_file(Run.statefile) && exist_file(Run.pidfile)) {
    
    return (get_timestamp(Run.pidfile, S_IFREG) >
	    get_timestamp(Run.statefile, S_IFREG));
    
  }

  return FALSE;
  
}


/**
 * Update the current process list with data from the state file. We
 * do *only* change processes found in *both* the monitrc file and in
 * the state file. The algorithm:
 *
 * Assume the control file was changed and a new process (B) was added
 * so the monitrc file now contains the processes: A B and C. The
 * running monit daemon only knows the processes A and C. Upon restart
 * after a crash the monit daemon first read the monitrc file and
 * creates the process list structure with A B and C. We then read the
 * state file and update the process A and C since they are found in
 * the state file, B is not found in this file and therefore not
 * changed.
 *
 * The same strategy is used if a process was removed, e.g. if the
 * process A was removed from monitrc; when reading the state file,
 * process A is not found in the current process list (the list is
 * always generated from monitrc) and therefore A is simply discarded.
 *
 * Finally, after the monit process state is updated this function
 * writes the new state file.
 */
void state_update() {

  int i;
  int l= 0;
  State_T s;
  FILE *S= NULL;
  Process_T new;
  int has_error= FALSE;
  
  if(! (S= open_state("r")))
      return;
  
  errno= 0;
  if(fread(&l, 1, sizeof (int), S) != sizeof(int)) {
    log("%s: Unable to read monit state information from '%s'\n",
	prog, Run.statefile);
    has_error= TRUE;
    goto error;
  }

  if(l > 0) {
    for(i=0; i<l; i++) {
      if(fread(&s, 1, sizeof(State_T), S) != sizeof(State_T)) {
	log("%s: An error occured when updating monit state information\n",
	    prog);
	has_error= TRUE;
	goto error;
      }
      if((new= get_process(s.name))) {
	update_process_state(new, &s);
      }
    }
  }

  error:
  close_state(S);

  if(!has_error)
      state_save();
  
}


/**
 * Finalize state
 */
void state_finalize() {
  
  unlink(Run.statefile);

}


/* ----------------------------------------------------------------- Private */


static FILE *open_state(const char *mode) {

  FILE *S= NULL;
  
  ASSERT(Run.statefile);

  umask(MYPIDMASK);
  
  if((S= fopen(Run.statefile, mode)) == NULL) {
    log("%s: Cannot open the monit state file '%s' -- %s\n",
	prog, Run.statefile, STRERROR);
    
    return NULL;
    
  }

  return S;
  
}


static void close_state(FILE *S) {

  ASSERT(fclose(S)==0);

}


static void clone_state(Process_T p, State_T *s) {
  
  memset(s, 0, sizeof(State_T));
  
  strncpy(s->name, p->name, STRLEN);
  s->mode= p->mode;
  s->nstart= p->nstart;
  s->ncycle= p->ncycle;
  s->do_validate= p->do_validate;
  s->has_checksum_error= p->has_checksum_error;

}


static void update_process_state(Process_T p, State_T *s) {

  p->mode= s->mode;
  p->nstart= s->nstart;
  p->ncycle= s->ncycle;
  p->do_validate= s->do_validate;
  p->has_checksum_error= s->has_checksum_error;
  
}
